/*
   $Id: GlintEngineFreq.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <iostream>
#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cerrno>
#include <vector>
#include <fstream>
#include <iomanip>
#include <cerrno>
#include <string>
#include <stdio.h>
#include "ptypes.h"

using namespace std;

#include "buffer.h"
#include "GlintEngineFreq.h"
#include "CompactTree.h"
#include "dust.h"


#include "Parameters.h"
#include "Bank.h"
#include "QueryBuffer.h"
#include "utils.h"

const int8_t SStrand[2]={1,-1};

GlintEngineFreq::GlintEngineFreq(QueryBuffer* q, Display& d): GlintEngine(q, d), frequentBuffer(NULL)
{
	CondemnationScore[ LOW ] = -(100/_parameters.GetConseqFreq())+1;
	CondemnationScore[ HIGH ] = 1;
	
	SuitableScore[ LOW ] = 1;
	SuitableScore[ HIGH ] = -(100/_parameters.GetConseqUnFreq())+1;
	SuitableScore[ CONDEMNED ] = -numeric_limits<double>::max();
}


/**
 * \brief Start the comparison between the sequence stored in the QueryBuffer and the bank singleton
 *         -# Mask low complexity sequences if asked ( -C > 0)
 *         -# Search for seeds on the plus and on the minus strands
 *         -# Align the seeds and create the hsps on the plus and on the minus strands
 */
void GlintEngineFreq::Run ()
{
	QueryBuffer* query = GetQueryPtr(); 

	__SearchingForSeeds();

	//-------------FREQ--------------

	IdentifySuitableZone();
	_display.SetQuery(query);
	_display.DisplaySuitableFasta(frequentBuffer, SuitableRegions);
	
	//cerr << "nb Suitable region : " <<(int)SuitableRegions.size()<< endl;
}



/**
 * \brief Start the comparison between the sequence stored in the QueryBuffer and the bank singleton
 *         -# Mask low complexity sequences if asked ( -C > 0)
 *         -# Search for seeds on the plus and on the minus strands
 *         -# Align the seeds and create the hsps on the plus and on the minus strands
 */
void GlintEngineFreq::BuildHsps ()
{
	QueryBuffer* query = GetQueryPtr(); 
	
	__MaskLowComplexity();
	__SearchingForSeeds();
	
	//-------------FREQ--------------
	
	IdentifySuitableZone();
	
	DisplayFreq* displayFreq_ptr = new DisplayFreq();
	DisplayFreq& displayFreq = *displayFreq_ptr;
	displayFreq.SetQuery(query);
	displayFreq.DisplaySuitableFasta(frequentBuffer, SuitableRegions);
	
	cerr << "nb Suitable region : " <<(int)SuitableRegions.size()<< endl;
}

/**
 *  \brief Compare a sequence against the bank
 *         - Using a sliding window, compare the sequence against the bank  
 * 		   - Create a boolean IsFrequent true when __nbTotalPos > countCutoff
 * 		   - Push the boolean in __frequentBuffer
 * 
 *  \param [in] seq_base The sequence address base
 *  \param [in] seq_length
 *  \param [out] match_results: used to put the retireved positions to the match pile managed by match_results
 * 
 * 	\private
*/
void GlintEngineFreq:: __SearchingForSeeds()
{
	size_t              seq_length = _query_ptr -> GetSeqLength();
	const unsigned char * seq_base = _query_ptr -> GetBase();
	
	frequence f;
	Window window(_mask);
	uint8_t msk_len = _mask.GetLength();
	
	size_t buffer_size = seq_length-msk_len;
	frequentBuffer     = new Buffer(buffer_size); //WARNING : TODO for multifasta delete the frequentBuffer after each sequence
	
	FILE * pFile;
	pFile = fopen ("pos.tsv","w");

	// The pos variable corresponds to the first base of the sequence (absolute position)
	for (uint32_t pos=1; pos <= seq_length; pos++)
	{
		char c=*(seq_base + pos);
		bool IsFrequent = true;

		uint32_t NbTotPos = 0;
		
		if (!Word::IsCode(c)) // Masked region
		{
			window.Reset();
		}
		else
		{
			window.Shift(c);        // The complexity is filtered out on the request
			if (window.IsFull()) 
			{
				if ( pos % _parameters.GetStep() == 0 )
				{
					_ctree.CompareWindow(window,pos-msk_len+1,__plus, *_query_ptr);
					NbTotPos += _ctree.GetNbPos();
					__plus.MatchReset();
					_ctree.CompareWindow(window,pos-msk_len+1,__minus, *_query_ptr);
					NbTotPos += _ctree.GetNbPos();
					__minus.MatchReset();
				}
				fprintf (pFile, "%s\t%d\n", window.GetWord().Convert().c_str(), NbTotPos);
				
				if ( NbTotPos < _parameters.GetCountCutoff() )
				{
					IsFrequent = false;
				}
				//match_results.Reset();
			}
		}
		f = (IsFrequent) ? HIGH : LOW;
		_BufferPut( f );
	}
	fclose (pFile);
}

/**
 * \brief _DisplayBuffer
* 		- display buffer content in fasta format 
 * 		_ inspired of QueryBuffer::GetSeqAsFasta().
 *
 * \param[in] The HitPile to print
 */

void GlintEngineFreq::_DisplayBuffer(const QueryBuffer& query)
{
	string      seqId = query.GetSeqId();
	size_t      size  = GetBufferSize();
	frequence * fbase = GetBufferBase();
	
	cout<<">"<< seqId.c_str() <<" countCutoff="<<(int)_parameters.GetCountCutoff()<<endl;
	
	uint32_t i;
	for ( i = 1 ; i <= size ; ++i )
	{
		frequence f = *(fbase+i);
		cout<<Decode(f);
		if ( i%60 == 0 ) cout<<" "<<i<<" "<<endl;
	}
	if ( i%60 != 0 ) cout<<endl<<endl;
	
}

/**
 * \brief _DisplayFasta
 * 		- display query sequence content in fasta format between 
 * 		the positions start and end .
 * 		inspired of QueryBuffer::GetSeqAsFasta().
 *
 * \param[in] query the query
 * \param[in] start the start position of the display
 * \param[in] end   the end position of the display
 */

void GlintEngineFreq::_DisplayFasta(const QueryBuffer& query, uint32_t start, uint32_t end)
{
	string    seqId   = query.GetSeqId();
	frequence * fbase = GetBufferBase();

	
	cout<<">"<< seqId.c_str() <<" suitable zone  start : "<<start<<" end : "<<end<<endl;
	
	uint32_t j = 0;
	uint32_t i = 0;
	for (j=start; j<end; ++j)
	{
		char a = *(fbase+j);
		char c = *(seq_base+j);
		if(a==LOW)
		{
		cout<<c;
			if((i+1)%60==0)
			{
				cout<<endl;
			}
		}
		else
		{
			cout<<'N';
			if((i+1)%60==0){cout<<endl;}
		}
		++i;
	}
	if ( i%60 != 0)
	{
		cout<<endl<<endl;
	}
}

/**
 * \brief _DisplaySuitableFasta
 * 		- display query sequence content in fasta format between 
 * 		the positions start and end .
 * 		inspired of QueryBuffer::GetSeqAsFasta().
 *
 * \param[in] query the query
 * \param[in] start the start position of the distplay
 * \param[in] end   the end position of the display
 */

void GlintEngineFreq::_DisplaySuitableFasta(const QueryBuffer& query)
{
	string seqId = query.GetSeqId();
	
	stack<regionFreq> regionTmp;
	regionTmp = SuitableRegions;
	regionFreq r;
	
	while (!regionTmp.empty())
		{
			r=regionTmp.top();
			_DisplayFasta(query, r.start, r.end);

			regionTmp.pop();
		}
		cout<<endl;
}

/**\brief _DisplayFastaATGCN
 * 		- display query sequence content in fasta format between 
 * 		the positions start and end .
 * 		inspired of QueryBuffer::GetSeqAsFasta().
 * 		- the frequent mers are represent by a 'N', 
 * 		thus we keep the information for design Probe
 *
 * \param[in] query the query
 */

void GlintEngineFreq::_DisplayFastaATGCN(const QueryBuffer& query)
{
	string	    seqId  = query.GetSeqId();
	size_t      size  = GetBufferSize();
	frequence * fbase  = GetBufferBase();
	
	cout<<">"<< seqId.c_str() <<endl;
	
	uint32_t j = 0;
	uint32_t i = 0;
	
	for (j=0; j<size; ++j)
	{
		char a = *(fbase+j);
		char c = *(seq_base+j);

		if(a==LOW)
		{	
			cout<<c;
			if((i+1)%60==0){cout<<endl;}
		}
		else
		{
			cout<<'N';
			if((i+1)%60==0){cout<<endl;}
		}
		++i;
	}
	if ( (i)%60 != 0)
	{cout<<endl<<endl;}
}


void GlintEngineFreq:: SetCondemnedZone() 
{
	frequence * fbase = GetBufferBase();
	regionFreq r;

	while (!CondemnedRegions.empty())
		{
			r=CondemnedRegions.top();
			for( uint32_t i = r.start ; i <= r.end ; ++i )
			{
				fbase[i] = CONDEMNED;
			}
			CondemnedRegions.pop();
			
		}
}


/** \brief Identify condemned zone using a scoring system
 * 
*/

void GlintEngineFreq:: IdentifyCondemnedZone() 
{
	//cerr<<"lancement de condemned  zone"<<endl;
		
	uint32_t      size             = GetBufferSize();
	frequence* fbase            = GetBufferBase();
	uint32_t* 	   S                = (uint32_t*) malloc( size*sizeof(uint32_t) );
	uint32_t      CondemnedMinSize = _parameters.GetCondemnedMinSize();
	
	uint32_t   start            = 0;
	S[0] =  CondemnationScore[*fbase];
	regionFreq r;
	
	
	if ( !_parameters.IsSilent() )
	{
		cerr << "CONDEMNED REGIONS\n=================" << endl;
	}
	for ( uint32_t i = 1  ; i < size ; ++i )
	{ 
		// S(i)= max (0 ,S(i-1) + score (l))
		frequence f =*( fbase+i );
		S[i] = ( S[ i-1 ] + CondemnationScore[ f ] > 0 ) ? S[ i-1 ] + CondemnationScore[ f ] : 0;
		if ( S[i] == 0 )
		{
			if ( S[ i-1 ] > 0 &&  ( i - start - 2) > CondemnedMinSize ) // real start = start + 1 et real end = i-1
			{
			  r.start =  start + 1;
			  r.end =  i - 1;
			  CondemnedRegions.push(r);
			  if ( !_parameters.IsSilent() )
			  {
				  cerr<<"start : "<<r.start<< " \tend : "<< r.end << "\tsize : "<< r.end-r.start<<endl;
			  }
			}			
			start = i;
		}

	}
	if ( !_parameters.IsSilent() )
	{
		cerr << "nb of Condemned regions : " <<(int)CondemnedRegions.size()<< endl << endl;
	}
	SetCondemnedZone();	
}		



/** \brief Identify suitable zones using a scoring system
 * 
 * \param [in] 
 * \param [out]
 * 
 * \public
*/

void GlintEngineFreq:: IdentifySuitableZone() 
{
	
	IdentifyCondemnedZone();
	
	uint32_t      size             = GetBufferSize();
	frequence* fbase            = GetBufferBase();	
	uint32_t* 	   S                = (uint32_t*) malloc( size*sizeof(uint32_t) );
	uint32_t      SuitableMinSize  = _parameters.GetSuitableMinSize();
	
	uint32_t start = 0;
	S[0] =  SuitableScore[*fbase];
	regionFreq r;
	
	if ( !_parameters.IsSilent() )
	{
		cerr << "SUITABLE REGIONS\n=================" << endl;
	}
	for ( uint32_t i = 1  ; i < size ; ++i )
	{ 
		// S(i)= max (0 ,S(i-1) + score (l))
		frequence f =*( fbase+i );
		S[i] = ( S[ i-1 ] + SuitableScore[ f ] > 0 ) ? ( S[ i-1 ] + SuitableScore[ f ] ) : 0;
		
		if ( S[i] == 0 || ( i == 1 && S[i] > 0 )  )
		{
			if ( S[ i-1 ] > 0 &&  (( i - start - 1) > SuitableMinSize )) // real start = start + 1 et real end = i-1
			{
				r.start =  start + 1;
				r.end =  i - 1;
				
				if ( !_parameters.IsSilent() )
				{
					cerr<<"start : "<<r.start << "\tpos : " <<(int)S[start+1]<< " \tend : "<< r.end<<"\tpos : " <<(int)S[i-1]<< "\tsize : "<< r.end-r.start<<endl;
				}
				SuitableRegions.push(r);
			}			
			start = i;	
		}
		
		if ( S[size-1] > 0 && (i - start -2) > SuitableMinSize ) // real start = start + 1 et real end = i-1
		{
			r.start =  start + 1;
			r.end =  i - 1;
			if ( !_parameters.IsSilent() )
			{
				cerr<<"start : "<<r.start<< " \tend : "<< r.end<< "\tsize : "<< r.end-r.start<<endl;
			}
			SuitableRegions.push(r);
		}			
	}
	if ( !_parameters.IsSilent() )
	{
		cerr << "nb of Suitable regions : " <<(int)SuitableRegions.size()<< endl<<endl;
	}
}
