/*
   $Id: IndexTree.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __INDEX_TREE_H__
#define __INDEX_TREE_H__

#include <stdio.h>
#include <limits>
#include <deque>

#include "ptypes.h"
#include "memory.h"
#include "window.h"
#include "buffer.h"
#include "Parameters.h"
#include "Bank.h"
#include "masque.h"


/**
 * \brief The template class SingleLinkedList defines a single linked list, used to temporary store objects like
 *        Node, Leaf, Pos (see later). 
 *        The objects are pushed through the front, and popped from the back
 *        You pop only One object, but you can push a single object, OR a whole list of objects
 * \pre The object must have a field brother, and SingleLinkedList<T> must be a friend of T
 */

template <class T> class SingleLinkedList {
    public:
    SingleLinkedList(): first(NULL),last(NULL) {};

    T * PopBack();
    void PushFrontList(T * p);
    void PushFront(T * p);

    bool IsEmpty() const {return first==NULL;};
    //bool IsEmpty() const {return true;};
    void Clear() { first = NULL; last = NULL;};

    private:
    T * first;
    T * last;
};

class Node;
class Leaf;
class Pos;
class IndexTree;

/**
 * \brief NBaseNode,NBaseLeaf and NBasePos are base classes for the Node, Leaf and Pos classes.
 *        The new operator is overloaded, so that the memory is taken:
 *           -From a "reserve" static object of class SingleLinkedList if the reserve is not empty
 *           -From the single instance of MemoryPool if the reserve is empty
 *        The delete operator is overloaded, so that the memory is not freed, but stored to the reserve
 *        Clear() clears the reserve, and MemoryUsed returns the quantitiy of memory taken from the pool
*/

class NBaseNode {
    public:
    static void * operator new (size_t sz);
    static void operator delete(void * l);
    static size_t MemoryUsed() { return mem_used;};
    static void Clear() { mem_used = 0; reserve.Clear(); };

    private:
    static size_t mem_used; /* The number of bytes in memory occupied by the positions */
    static SingleLinkedList<Node> reserve;
};


class NBaseLeaf {
    public:

    static void * operator new (size_t sz);
    static void operator delete(void * l);
    static size_t MemoryUsed() { return mem_used;};
    static void Clear() { mem_used = 0; reserve.Clear(); };

    private:
    static size_t mem_used; /* The number of bytes in memory occupied by the positions */
    static SingleLinkedList<Leaf> reserve;
};

class NBasePos {
    public:
    /* WARNING - Here, the delete operator pushes a whole chain of elements, thus allowing for the deletion of
    many Pos objects together - Used by the destructor od Leaf
    */
    static void * operator new (size_t sz);
    static void operator delete(void * l);
    static size_t MemoryUsed() { return mem_used;};
    static void Clear() { mem_used = 0; reserve.Clear(); };

    private:
    static size_t mem_used; /* The number of bytes in memory occupied by the positions */
    static SingleLinkedList<Pos> reserve;
};

/**
 * \brief The class Node defines a node in the indexation tree (see the object IndexTree).\n
 *        The new and delete operators are overloaded for this class, thus allowing for very fast memory allocation.
 *        From time to time, the IndexTree object will free the whole memory pool, thus destroying all Node objects altogether.\n
 *        IndexTree and Leaf are friends of Node, the IndexTree may set/get the brother and hash_code private members
 */
class Node: public NBaseNode {
    public:

/**
 * \brief Default constructor, for initializing the array top
 *        hash_code is not used here, the index is the hash_code
 */
	    Node();
    
/**
 * \brief This constructor is used to insert a new node inside an horizontal branch
 * \pre (not checked): prev_brother != NULL
 */
	    Node(uint32_t code, Node * prev_brother);
    
/**
 * \brief This constructor is used to insert a new node at front of an horizontal branch
 * \pre (not checked): father != NULL
 * \param Last int parameter is dummy, only used to correctly overload the function
 */
	    Node(uint32_t code, Node* father, int);
	
	    /* Printing a node */
	    void Print(ostream&, const Mask&, uint8_t level=ROOT_LEVEL, const string& word="", int indent=0);
	
	    friend class IndexTree;
	    friend class Leaf;
	    friend class SingleLinkedList<Node>;
	
    private:
		uint32_t hash_code;	/* The hash code, at the correct level, of the words managed by this node */
		Node* brother;      /* Next node (same level)   */
	    void* son;          /* Next node (level + 1) - The node may be a Node or a Leaf object */
};

class Pos : public NBasePos {
    public:

    /* The constructor of Pos */

    Pos(uint32_t p):position(p),brother(NULL){};

    friend class Leaf;
    friend class IndexTree;
    friend class SingleLinkedList<Pos>;

    private:
    uint32_t position; /* The position to keep */
    Pos* brother;   /* Next position (same leaf) */
};


/**
 * \brief The class Leaf defines a leaf in the indexation tree (see the object IndexTree).\n
 *        The new and delete operators are overloaded for this class, allowing for very fast memory allocation.
 *        From time to time, the IndexTree object will free the whole memory pool, destroying all Leaf objects altogether.
 *        Each leaf contains an array of positions: the positions of the word corresponding to the leaf.\n
 *        IndexTree is friend of Leaf, thus IndexTree may set/get the brother and hash_code private members
*/

class Leaf : public NBaseLeaf {
    public:

/**
 * \brief Used to insert a new leaf inside an horizontal branch
 * \pre (not tested): prev_brother != NULL
 * \param code The code
 * \param prev_brother The position where the leaf will be inserted
 */
    Leaf(uint32_t code, Leaf* prev_brother);
    
/**
 * \brief Used to insert a new leaf at front of an horizontal branch
 * \pre (not tested): father != NULL
 * \param code The code
 * \param father The upper node
 * \param unknown A dummy parameter, to correctly overload the function
 */
     
    Leaf(uint32_t code, Node* father,int);

    /**
     * \brief Leaf has a destructor, because when a leaf is destroyed we have to delete the Positions it encapsulates.
     *        However, we delete ONLY the first pos object, because the delete operator is overloaded for the Pos class,
     *        so that the whole linked list will be recycled (see the Pos class)
    */
    ~Leaf() {
        if (pos != NULL)
            delete pos;
    };

    /* Some accessors */
    int   GetNbPos() const {return nb_pos;};
    bool IsEmpty() const {return !(nb_pos);};

    /* Printing a leaf */
    void Print(ostream&, const Mask&, const string& word="", int indent=0);

    /* Storing positions */
    uint8_t NewPos(uint32_t,uint8_t);

    friend class IndexTree;
    friend class SingleLinkedList<Leaf>;

    private:
    uint32_t hash_code;	/* The hash code, at leaf level, of the word */
    Leaf* brother;      /* Next leaf (same level) */
	Pos*  pos;			/* pointeur sur les positions   */
    uint8_t nb_pos;       /* Number of cells in *pos array. Should be == nb_pos, but is not always */
};


/**
 * \brief The class BaseSerializeShift is an abstract functor, used only through the () operator.\n
 *        Its derived classes (the template SerializeShift) define the operator(), writing the argument shift
 *        to the argument pf (1st version) or to the MemArray m (2nd version)\n
 *        A type cast is first performed, thus leading to 1, 2 or 4 bytes written to the file, depending upon T
 *        Note that this should work ONLY for the uint8_t, uint16_t, uint32_t data types.\n
 *        The SerializeShift objects are stored in a private array in the class IndexTree, this array
 *        is initialized by the constructor of IndexTree.
*/

class BaseSerializeShift {

	public:
	virtual ~BaseSerializeShift() {}
    virtual uint8_t operator() (FILE* pf, size_t shift) = 0;
    virtual uint8_t operator() (Buffer& m, size_t shift) = 0;

};

template<class T> class SerializeShift: public BaseSerializeShift {
    public:
    virtual uint8_t operator() (FILE* pf, size_t shift) {
        T s = (T) shift;
        if (shift>limit.max())
            return 0;
        else
            return sizeof(T)*fwrite(&s,sizeof(T), 1, pf);
    };
    virtual uint8_t operator() (Buffer& m, size_t shift) {
        T s = (T) shift;
        if (shift>limit.max())
            return 0;
        else
            return sizeof(T) * (uint8_t) m.Put(s);    // m.Put returns true/flase, casted to 1/0
    };
    private:
    numeric_limits<T> limit;
};

/**
 * \brief The class IndexTree defines an indexation tree. The tree manages:
 *            - Top nodes, stored in an array of length 2*n where n is the length of the ROOT_LEVEL (see the Mask class)
 *            - Mask.GetLevel(leaf_level)-2 levels of nodes, the number of levels thus depends on the mask
 *            - A level of leafs.\n
 *        IndexTree handles a 'pass' mechanism, in which only some slices of the top nodes (thus of the tree) are considered for inserting,
 *        searching of deleting the objects. This mechanism enables the indexation of very large files, with a lot of words and a lot of positions,
 *        without using too much memory.\n
 *        The memory pools are flushed at the end of each pass, thus destroying all nodes and leaves.
*/

class IndexTree : public Singleton<IndexTree> {
    private:
	    IndexTree();
	    ~IndexTree();
	    
	public:
		friend class Singleton<IndexTree>;

	    void SetNbPass(uint8_t);         // Setting the number of passes for creating and writing the index files
	    uint8_t NextPass();
	    uint8_t GetPass() { return ct_pass;};
	    uint8_t GetNbPass() { return nb_pass;};
	    bool  Phylum(const Window &w) const;          // Return true if the word is in the good interval, false if not

	    Leaf* Insert(const Window & w);
	    Leaf* Search(const Window& w);
	    int   Delete(const Window &w);
	
	    /* Printing or binary writing the tree */
	    friend ostream & operator<<(ostream&, const IndexTree&);
	    string GetInfo();
	
	    int WriteCompressed(FILE * pf);
	    int WriteCompressed(Buffer*);

    private:
	    const Parameters& __parameters;
	    Bank& __bank;
	    const Mask& __mask;
	    Node* top;      /* The top of the tree: an array of Node */
	    uint32_t nb_top;   /* The number of phyla */
	    Bank* bank;
	    int leaf_level; /* The result of bank->GetMsk().GetLeafLevel() */
	
	    /* A stack to temporary store the info about the top nodes, used by WriteCompressed */
	    struct TopInfo {   // Some info to store in a stack about the top nodes
	        TopInfo():sons(0),start(0){};
	        uint8_t sons;    // The nb of sons of this top node
	        size_t start;   // The start position of this top node
	    };
	    TopInfo enul;      // A null element
	    deque<TopInfo> top_info;
	
	    /* The offset, used by WriteCompressed */
	    size_t offset;  // Used as a temporary storage by the __Write functions
	
	    /* The shift written in the binary file, may be written by several functions */
	    BaseSerializeShift* serialize_shift[10];
	
	    /* Managing several passes */
	    uint8_t nb_pass;  // Number of passes
	    uint8_t ct_pass;  // Current pass number
	    uint32_t lg_pass;  // The "length" of a pass
	    uint32_t inf_code; // For a given pass, we consider only the ROOT_LEVEL codes in the interval [inf_code,sup_code[
	    uint32_t sup_code;
	
	    void __FreeAll();   // Freeing all memory in one call
	
	    Leaf*  __Search(Node *, const Window&, uint8_t);
	    Node * __HorizNodeSearch(Node *, uint32_t);
	    Node * __HorizNodeSearch(Node *, uint32_t, Node **);
	    Leaf*  __HorizLeafSearch(Leaf *, uint32_t);
	    Leaf*  __HorizLeafSearch(Leaf *, uint32_t, Leaf ** prev);
	
	    int  __Delete(Node *, const Window& w, uint8_t level);
	
	    /* Low level functions to write data to a file */
	    uint8_t __WriteSons(FILE*, void*,uint8_t);
	    void  __SerializeNode(FILE *,Node*, uint8_t, size_t, uint8_t);
	    void __SerializeLeaf(FILE *, Leaf*, size_t, uint8_t);
	    int __WriteLeafPositions(FILE*, Leaf*);
	    int __CheckAndWriteOffset(FILE*,size_t ,uint8_t );
	
	    /* Low level functions to write the data to a memory array */
	    uint8_t __WriteSons(Buffer&, void*,uint8_t);
	    void  __SerializeNode(Buffer&,Node*, uint8_t, size_t, uint8_t);
	    void __SerializeLeaf(Buffer&, Leaf*, size_t, uint8_t);
	    int __WriteLeafPositions(Buffer&, Leaf*);
		int __CheckAndWriteOffset(Buffer&,size_t ,uint8_t );
	
};


/* the inline methods of the classes NBaseNode, NBaseLeaf, NBasePos, Node, Leaf */

inline void * NBaseNode::operator new (size_t sz)
{
    if (reserve.IsEmpty())
    {
        mem_used += sz;
        return MemoryPool::Instance().Malloc(sz);
    }
    else
    {
        return (void *) reserve.PopBack();
    };
}
inline void NBaseNode::operator delete(void * l)
{
    reserve.PushFront((Node*)l);
}

inline void * NBaseLeaf::operator new (size_t sz)
{
    if (reserve.IsEmpty())
    {
        mem_used += sz;
        return MemoryPool::Instance().Malloc(sz);
    }
    else
    {
        return (void *) reserve.PopBack();
    };
}
inline void NBaseLeaf::operator delete(void * l)
{
    reserve.PushFront((Leaf*)l);
}

inline void * NBasePos::operator new (size_t sz)
{
    if (reserve.IsEmpty())
    {
        mem_used += sz;
        return MemoryPool::Instance().Malloc(sz);
    }
    else
    {
        //cerr << "coucou recyclage\n";
        return (void *) reserve.PopBack();
    };
}
inline void NBasePos::operator delete(void * l)
{
    //cerr << "coucou delete\n";
    reserve.PushFrontList((Pos*)l);
}

inline Node::Node() : hash_code(0),brother(NULL),son(NULL) {}
inline Node::Node(uint32_t code, Node * prev_brother):hash_code(code),brother(prev_brother->brother),son(NULL)
{
    prev_brother->brother = this;
}
inline Node::Node(uint32_t code, Node* father, int):hash_code(code),brother((Node*)father->son),son(NULL)
{
    father->son = this;    // father must be != 0
}

inline Leaf::Leaf(uint32_t code, Leaf* prev_brother):hash_code(code),brother(prev_brother->brother),pos(NULL),nb_pos(0)
{
    prev_brother->brother = this;
}

inline Leaf::Leaf(uint32_t code, Node* father,int):hash_code(code),brother((Leaf*)father->son),pos(NULL),nb_pos(0)
{
    father->son=this;    // father must be != 0
}
#endif
