/*
   $Id: memory.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _MEMORY_H
#define _MEMORY_H

#include <vector>
using namespace std;

#include "ptypes.h"
#include "utils.h"

/** \brief the MemoryPool class implements a very simplistic object used for allocating/releasing memory.
 *         - When needed, the memory is allocated by pools of 100 Mb
 *         - Malloc reserves a few bytes on top of current pool, creating a new pool if necessary.
 *         - Free does nothing, thus organizing massive memory leaks...
 *         - FreeAll frees all the pools: memory allocation/deallocation is thus very fast, but not very flexible.
 *         - MemoryPool is a singleton
*/

class MemoryPool: public Singleton<MemoryPool> {
    public:
	    void* Malloc(size_t);
	    void Free(void*ptr) {};
	    void FreeAll();
	    size_t MemoryUsed() { return mem_used;};
	    size_t MemoryAllocated() { return mem_alloc;};
	    friend class Singleton<MemoryPool>;

    private:
	    MemoryPool():pool_size(100*1048576),mem_used(0),mem_alloc(pool_size),current_pool(0) //Size of a pool = 100Mb
	    {
	        void* base = MAlloc(pool_size);
	        pool.push_back((uint8_t*)base);
	        pool_top.push_back((uint8_t*)base);
	    };
	    ~MemoryPool()
	    {
			for (vector<uint8_t*>::iterator i = pool.begin(); i != pool.end(); ++i)
			{
				free ( *i );
			}
		}

	    const size_t pool_size;
	    size_t mem_used,mem_alloc;  // mem_used = memory used so far. mem_alloc = higher than mem_use, memory given by malloc calls
	    int current_pool;
	    vector<uint8_t*>pool;         // The base address of each pool
	    vector<uint8_t*>pool_top;     // The top of each pool
};

#ifdef REMOVED
/*
     The function MaxSize tries to determine the size of the memory (RAM, not virtual) of this computer
     It returns the number of bytes we may probably use
     We call malloc/free to really retrieve the wanted memory
     Parameters:
          -If 0, returns 80% of the total memory
          -If > 0, specifies the fraction of memory we are allowed to use (50 = the program may use half of the available memory).
     Retun value: The size of allocatable memory

*/
size_t MaxSize(int);
#endif

#endif
