/*
   $Id: window.h 650 2011-10-31 14:25:33Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _WINDOW_H
#define _WINDOW_H

#include "ptypes.h"
#include "word.h"

class Window: private NonCopyable {
    public:
    /* The constructor - Parameter = The StringMask, from which the length of the window (number of characters) is computed */
    Window(const Mask& msk);
    /*
    Shifting or resetting the window:
       -Shift shifts the window
       -Reset resets the window
       -ResetComp resets the window AND the comp object
    */

    void Shift(char c);
    void Reset();

    /* Checking the window: is it full or not ? is there enough complexity or not ? What is the length ?*/
    bool IsFull() const{ return (pos==size);};
    bool IsEmpty() const {return pos==0;};
    uint8_t GetLength() const {return length;};
    uint8_t GetSize() const {return size;};
    uint8_t GetPos() const {return pos;};

    /* Retrieve the Word and the mask, but only for using the accessor functions (the objects retrieved are const) */
    const Word& GetWord() const {return w;};

    /*
        The hashing functions used to store and retrieve the "words" in the tree:

        -TreeHash        ==> Used in conjunction with the mask for storing and retrieving entries in the tree, based upon the word
                             handled by the window
        -ReverseTreeHash ==> Same as TreeHash, but for the reverse complement word

    */

    uint32_t TreeHash(uint8_t level) const;
    uint32_t ReverseTreeHash(uint8_t level) const;

    private:

    const Mask& m;             /* The mask, used by the hashing functions */
    Word w;             /* The word encapsulated by the window */
    uint8_t length;         /* Length in characters */
    uint8_t size;           /* Size (in bits) */
    int pos;            /* The position of the window. In bits */

    mutable uint32_t tree_hash[MAX_LEAF_LEVEL+1];     // Caching the results of ThreeHash for all the levels
    mutable bool valid_tree_hash;
    mutable uint32_t rev_tree_hash[MAX_LEAF_LEVEL+1]; // Caching the results of ReverseThreeHash for all the levels
    mutable bool valid_rev_tree_hash;
};


/* THE inline METHODS OF THE OBJECT: Window */

//inline Window::Window(const Mask& msk): m(msk),w(msk.GetLength()),
//length(msk.GetLength()),size(2*msk.GetLength()),pos(0),comp(4),
//valid_tree_hash(false),valid_rev_tree_hash(false)
inline Window::Window(const Mask& msk): m(msk),w(msk.GetLength()),
length(msk.GetLength()),size(2*msk.GetLength()),pos(0),
valid_tree_hash(false),valid_rev_tree_hash(false)
{
    for(int i =0; i<=MAX_LEAF_LEVEL;i++)
    {
        tree_hash[i]=0;rev_tree_hash[i]=0;
    };
}

inline void Window::Shift(char c) {
    valid_tree_hash     = false;
    valid_rev_tree_hash = false;
    w.Pushc(c);
    if (!IsFull())
        pos += 2;
}

inline void Window::Reset() {
    pos=0;
    //comp.Reset();
    valid_tree_hash=false;
    valid_rev_tree_hash=false;
}   // Clear position and reset the Complexity tool

/* There is a dependency between Window and Word (Window is a friend of Word)
   We have 2 versions for the hashing functions
*/

inline uint32_t Window::TreeHash(uint8_t level) const
{
    if (valid_tree_hash) return tree_hash[level];
    uint64_t r    = w.r;
    int count[MAX_LEAF_LEVEL+1];
    for (int i=0; i<=m.leaf_level; i++) {
        tree_hash[i]=0;
        count[i]=0;
    };

    for (int i=length-1; i>=0; i--)
    {
        int level = m.pos[i];      // The level of the mask at this position
        int c=r&3;                 // keep the last 2 bits of r
        r>>=2;                     // rotate r
        c <<= 2*count[level]++;    // shift c, incrementing the count
        tree_hash[level] |= c;     // Update the hash for this level
    };
    valid_tree_hash = true;

    /*JUST FOR DEBUG
    cerr << w.Convert() << '(' << (int)level << ")  ";
    for (int i=1; i<=m.leaf_level; i++)
    {
        cerr << ' ' << hex << tree_hash[i];
    }
    cerr << '\n';
    *//*
    cerr << w.Convert() << '*';
    for (int i=1; i<=m.leaf_level; i++)
    {
        cerr << ' ' << hex << TreeHash_OLD(i);
    }
    cerr << '\n' << '\n';

    ReverseTreeHash(level);

    cerr << w.Convert() << ' ' << (int)level << ' ' << hex << tree_hash[level] << '\n';
    */
    return tree_hash[level];       // return the hash at the good level
}

inline uint32_t Window::ReverseTreeHash(uint8_t level) const
{
    if (valid_rev_tree_hash) return rev_tree_hash[level];
    uint64_t r   = ~w.r;
    uint64_t msk = (((uint64_t)3)<<(w.size-2));

    int count[MAX_LEAF_LEVEL];
    for (int i=0; i<=m.leaf_level; i++) {
        rev_tree_hash[i]=0;
        count[i]=0;
    };


    for (int i=0; i<length; i++)
    {
        int level = m.pos[length-i-1];      // The level of the mask at this position
        uint64_t c=((r&msk)>>(w.size-2))&3; // keep the first 2 useful bits of ~r
        r<<=2;                     // rotate r
        c <<= 2*count[level]++;    // shift c, incrementing the count
        rev_tree_hash[level] |= c; // Update the hash for this level
    };
    valid_rev_tree_hash = true;

    /*JUST FOR DEBUG
    cerr << w.ReverseComplement()->Convert() << '(' << (int)level << "R) ";
    for (int i=0; i<=MAX_LEAF_LEVEL; i++)
    {
        cerr << ' ' << hex << rev_tree_hash[i];
    }
    cerr << '\n';
    */
    /*cerr << w.ReverseComplement()->Convert() << '!';
    for (int i=1; i<=m.leaf_level; i++)
    {
        cerr << ' ' << hex << ReverseTreeHash_OLD(i);
    }
    cerr << '\n' << '\n';
    */

    return rev_tree_hash[level];       // return the hash at the good level
}

#endif    // WINDOW_H

